using System;
using System.Collections.Generic;
using System.Reflection;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Components {
  [Tracked]
  public class WorldCamera : Component {
    private static bool HooksLoaded = false;
    private static IDetour OnSpriteBatchPushSpriteHook;

    public static void LazyLoad() {
      WorldCamera.HooksLoaded = true;

      On.Celeste.GameplayRenderer.Render += OnGameplayRendererRender;
      IL.Monocle.EntityList.RenderExcept += ILEntityListRenderExcept;

      OnSpriteBatchPushSpriteHook = new Hook(
        typeof(SpriteBatch).GetMethod("PushSprite", BindingFlags.NonPublic | BindingFlags.Instance),
        typeof(WorldCamera).GetMethod("OnSpriteBatchPushSprite", BindingFlags.NonPublic | BindingFlags.Static)
      );

      // render compat
      // On.Celeste.CrystalStaticSpinner.InView += OnCrystalSpinnerInView;
    }

    [OnUnload]
    public static void Unload() {
      WorldCamera.HooksLoaded = false;

      On.Celeste.GameplayRenderer.Render -= OnGameplayRendererRender;
      IL.Monocle.EntityList.RenderExcept -= ILEntityListRenderExcept;

      OnSpriteBatchPushSpriteHook?.Dispose();

      // render compat
      // On.Celeste.CrystalStaticSpinner.InView -= OnCrystalSpinnerInView;
    }


    public Hitbox Target;
    public Vector2 Position;

    internal float Width { get { return this.Target.Width; } }
    internal float Height { get { return this.Target.Height; } }

    internal Vector2 RealPosition { get { return base.Entity.Position + this.Position; } }
    internal Vector2 Delta { get { return this.Target.Position - this.RealPosition; } }

    private bool TransitivelyVisible = false;


    public WorldCamera(Vector2 position, Vector2 target, float width, float height) : this(
      position, new Hitbox(width, height, target.X, target.Y)
    ) {}

    public WorldCamera(Vector2 position, Hitbox target) : base(true, true) {
      this.Position = position;
      this.Target = target;
      if (!WorldCamera.HooksLoaded) { WorldCamera.LazyLoad(); }
    }


    public override void Update() {
      var level = this.Scene as Level;

      var padding = 16f;
      var renderBounds = new Hitbox(
        this.Target.Width + padding * 2, this.Target.Height + padding * 2,
        this.RealPosition.X - padding, this.RealPosition.Y - padding
      );
      var camera = level.Camera;

      this.Visible = true;
      // todo: this is broken :(
      // if (renderBounds.Intersects(camera.X, camera.Y, camera.Viewport.Width, camera.Viewport.Height)) {
      //   this.Visible = true;
      //   foreach (WorldCamera other in level.Tracker.GetComponents<WorldCamera>()) {
      //     if (
      //       !other.Visible
      //       && this.Target.Intersects(other.RealPosition.X, other.RealPosition.Y, other.Width, other.Height)
      //     ) {
      //       // todo: we need a way to override Visible on other to be always true
      //       // fixme: now we need a way to disable this dumbass :(
      //       other.TransitivelyVisible = true;
      //     }
      //   }
      // } else {
      //   this.Visible = false;
      // }
    }


    private void RenderEntityWith(Entity entity) {
      if (!(
        base.Active && (base.Visible || this.TransitivelyVisible)
        && entity.Visible
        && entity.Scene is Level level
      )) { return; }

      level.Camera.Position += this.Delta;

      try {
        SpriteBatchCull = this.Target;
        SpriteBatchDelta = this.Delta;

        // fixme: these methods are never called
        // entity compat
        if (entity is CrystalStaticSpinner spinner) { spinner.Update(); }
        else if (entity is DustStaticSpinner dust) { dust.Update(); }
        else if (entity is Lightning lightning) { lightning.ToggleCheck(); }

        entity.Render();
      } finally {
        level.Camera.Position -= this.Delta;
        SpriteBatchCull = null;
      }
    }


    private static bool IsRenderingGameplay = false;
    private static void OnGameplayRendererRender(
      On.Celeste.GameplayRenderer.orig_Render orig, GameplayRenderer renderer, Scene scene
    ) {
      IsRenderingGameplay = true;
      try {
        orig(renderer, scene);
      } finally {
        IsRenderingGameplay = false;
      }
    }


    private static void ILEntityListRenderExcept(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.Before, instr => instr.MatchCallOrCallvirt<Entity>("Render"))) {
        Logger.Log(LogLevel.Error, "bitsbolts.Components.WorldCamera", $"IL@{cursor.Index}: Hook failed to find entity.Render() in EntityList.RenderExcept!");
        return;
      }

      cursor.Emit(OpCodes.Dup);
      cursor.GotoNext(MoveType.After, instr => instr.MatchCallOrCallvirt<Entity>("Render"));
      cursor.EmitDelegate<Action<Entity>>(
        (entity) => {
          if (!(IsRenderingGameplay && entity.Scene is Level level)) { return; }

          foreach (WorldCamera camera in level.Tracker.GetComponents<WorldCamera>()) {
            camera.RenderEntityWith(entity);
          }
        }
      );
    }


    private static Hitbox SpriteBatchCull = null;
    private static Vector2 SpriteBatchDelta;
    delegate void SpriteBatchPushSprite_orig(
      SpriteBatch batch,
      Texture2D texture,
      float sourceX, float sourceY, float sourceW, float sourceH,
      float destinationX, float destinationY, float destinationW, float destinationH,
      Color color,
      float originX, float originY, float rotationSin, float rotationCos,
      float depth,
      byte effects
    );

    private static void OnSpriteBatchPushSprite(
      SpriteBatchPushSprite_orig orig,
      SpriteBatch batch,
      Texture2D texture,
      float sourceX, float sourceY, float sourceW, float sourceH,
      float destinationX, float destinationY, float destinationW, float destinationH,
      Color color,
      float originX, float originY, float rotationSin, float rotationCos,
      float depth,
      byte effects
    ) {
      if (SpriteBatchCull == null) {
        orig(
          batch,
          texture,
          sourceX, sourceY, sourceW, sourceH,
          destinationX, destinationY, destinationW, destinationH,
          color,
          originX, originY, rotationSin, rotationCos,
          depth,
          effects
        );
      } else {
        var (x, y, w, h) = Util.NormalizeRect(destinationX, destinationY, destinationW, destinationH);
        if (!SpriteBatchCull.Intersects(x, y, w, h)) { return; }

        orig(
          batch,
          texture,
          sourceX, sourceY, sourceW, sourceH,
          destinationX - SpriteBatchDelta.X, destinationY - SpriteBatchDelta.Y, destinationW, destinationH,
          color,
          originX, originY, rotationSin, rotationCos,
          depth,
          effects
        );
      }
    }
  }
}
