using System;
using System.Linq;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;


namespace Bitsbolts.Entities {
  [MapDataTracked("Bitsbolts/LoaderRespawn")]
  [CustomEntity("Bitsbolts/LoaderRespawn")]
  public class LoaderRespawn : Entity {
    [OnLoad]
    public static void Load() {
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;
      On.Celeste.Session.Restart += OnSessionRestart;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
      On.Celeste.Session.Restart -= OnSessionRestart;
    }


    public override void Added(Scene scene) {
      // only the entity data of this entity is used, not the entity itself, so we can remove it.
      base.Added(scene);
      base.RemoveSelf();
    }


    private static bool lastExitWasGolden = false;


    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level,
      Player.IntroTypes playerIntro, bool isFromLoader
    ) {
      // not loading from a new or resumed session
      if (!(playerIntro != Player.IntroTypes.Transition && isFromLoader)) {
        lastExitWasGolden = false;  // wtf?
        orig(level, playerIntro, isFromLoader);
        return;
      }

      // Since we haven't loaded any entities yet, custom spawn entities haven't
      // set their spawnpoint atm, so when loading into a room with no
      // Celeste.Spawns, this will be null, which we need to handle.
      // If there is actually no spawns, then this entire trigger won't work
      // entirely as expected but that's user error at that point.
      // todo: maybe throw up a postcard for this error? idk from where tho
      Vector2 oldSpawn = level.Session.RespawnPoint
        ?? new Vector2(Single.NegativeInfinity, Single.NegativeInfinity);

      // This differs from Session.JustStarted in that this is only triggered
      // the first time the map is loaded on a fresh session
      var newLevel = !level.Session.LevelFlags.Contains(level.Session.LevelData.Name);

      var spawn = level.Session.MapData.GetTracker()
        .GetEntityDatas("Bitsbolts/LoaderRespawn")
        .Where(data => {
          var onAreaStart = data.Bool("onAreaStart", true);
          var onSessionResume = data.Bool("onSessionResume", true);
          var onGoldenRestart = data.Bool("onGoldenRestart", true);
          var isGlobal = data.Bool("global", false);

          return (data.Level.Name == level.Session.Level || isGlobal)
            && (
              (onAreaStart && newLevel)
              || (onSessionResume && !newLevel)
              || (onGoldenRestart && lastExitWasGolden)
            );
        }).OrderBy(data => {
          // prefer global spawns first, goldenrestart second preference
          var isGlobal = data.Bool("global", false);
          var onGoldenRestart = data.Bool("onGoldenRestart", true);
          return (isGlobal ? -2 : 0) + (onGoldenRestart && lastExitWasGolden ? -1 : 0);
        }).ThenBy(
          // find the respawn closest to this entity
          data => Vector2.DistanceSquared(
            data.Position + level.Session.LevelData.Position,
            oldSpawn
          )
        ).FirstOrDefault();

      if (spawn == null) {
        // no loader respawn exists or is applicable, stop
        lastExitWasGolden = false;
        orig(level, playerIntro, isFromLoader);
        return;
      }

      level.Session.RespawnPoint = spawn.Position + spawn.Level.Position;
      level.Session.Level = spawn.Level.Name;
      // setting StartedFromBeginning to true when restarting from session
      // causes weird behavior
      level.Session.StartedFromBeginning = !spawn.Bool("taintSession", false)
        && !(spawn.Bool("onSessionResume", true) && !newLevel);

      lastExitWasGolden = false;
      orig(level, playerIntro, isFromLoader);

      if (!spawn.Bool("saveRespawnPoint", true)) {
        // and reset the spawn so the player instead uses the old spawn
        level.Session.RespawnPoint = oldSpawn;
      }
    }


    private static Session OnSessionRestart(
      On.Celeste.Session.orig_Restart orig, Session self, string intoLevel
    ) {
      Session restartSession = orig(self, intoLevel);

      if (
        intoLevel != null && Engine.Scene is LevelExit exit
        && DynamicData.For(exit).Get<LevelExit.Mode>("mode") == LevelExit.Mode.GoldenBerryRestart
      ) {
        lastExitWasGolden = true;
      }

      return restartSession;
    }
  }
}
