using System;
using System.Linq;
using System.Collections.Generic;
using System.Reflection;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;


using Bitsbolts.Entities;


namespace Bitsbolts.Hooks {
  public static class GlobalRoomLoader {
    private static IDetour ILLevelLoadLevelHook;

    [OnLoad]
    public static void Load() {
      Bitsbolts.Events.OnMapDataLoad += OnMapDataLoad;

      // todo: consider hooking (after bumping up everest dependency)
      // Celeste.Mod.Everest.Events.AssetReload.OnReloadLevel += InvalidateRoomCache;
      // Celeste.Mod.Everest.Events.AssetReload.OnReloadAllMaps += InvalidateRoomCache;

      ILLevelLoadLevelHook = new ILHook(
        typeof(Level).GetMethod("orig_LoadLevel", BindingFlags.Public | BindingFlags.Instance),
        ILLevelLoadLevel
      );
    }

    [OnUnload]
    public static void Unload() {
      ILLevelLoadLevelHook?.Dispose();
    }


    internal static String GLOBAL_ROOM_PREFIX = "_bb_global";
    internal static Vector2 GLOBAL_ROOM_OFFSET = new Vector2(-1e6f, 1e6f);
    internal static Vector2 GLOBAL_ROOM_TILE = new Vector2(0f, 1e4f);

    private static IEnumerable<LevelData> RoomIter = null;
    private static List<EntityData> EntityListBuffer = new List<EntityData>();


    private static void OnMapDataLoad(MapData mapData) {
      var levelControllers = GenerateLevelControllers(mapData);
      OffsetRoomEntities(levelControllers.Keys);

      DynamicData.For(mapData).Set("_bitsbolts_globalroom_levelcontrollers", levelControllers);
    }

    private static Dictionary<LevelData, GlobalRoomLoaderController> GenerateLevelControllers(MapData mapData) {
      Dictionary<LevelData, GlobalRoomLoaderController> levelControllers = mapData.Levels.Where(
        level => level.Name.StartsWith(GlobalRoomLoader.GLOBAL_ROOM_PREFIX)
      ).ToDictionary(level => level, _ => GlobalRoomLoaderController.DEFAULT_INSTANCE);

      foreach (EntityData controllerData in mapData.GetTracker().GetEntityDatas("Bitsbolts/GlobalRoomLoaderController")) {
        if (!levelControllers.ContainsKey(controllerData.Level)) {
          throw new Exception($"Global room controller with eid {controllerData.Level.Name}:{controllerData.ID} in map sid {mapData.Area.SID} is not placed in a global room; room name must have a prefix of {GLOBAL_ROOM_PREFIX}");
        } else if (levelControllers[controllerData.Level] == GlobalRoomLoaderController.DEFAULT_INSTANCE) {
          levelControllers[controllerData.Level] = new GlobalRoomLoaderController(controllerData, mapData);
        } else {
          throw new Exception($"Duplicate global room controller with eid {controllerData.Level.Name}:{controllerData.ID} in level {controllerData.Level.Name}!");
        }
      }

      return levelControllers;
    }

    // Hack: because the mess below, all global rooms are currently spawned at
    // with the same offset as the current room. Therefore, entities in global
    // rooms may spawn at the same position as those in other positions.
    // So for now, just move them out of the way, somewhere far away...
    private static void OffsetRoomEntities(IEnumerable<LevelData> levels) {
      Logger.Log(LogLevel.Debug, "bitsbolts.Hooks.GlobalRoomLoader", $"Offsetting rooms:");

      foreach (var pair in levels.Enumerate()) {
        var (index, room) = pair;
        var dd = DynamicData.For(room);

        // hack: DynamicData hates working with value types, so box the bool
        if ((bool) (dd.Get<bool?>("_bitsbolts_globalroomoffset") ?? false)) {
          continue;
        }

        var offset = GLOBAL_ROOM_OFFSET + GLOBAL_ROOM_TILE * index;
        foreach (var edata in room.Entities) {
          edata.Position += offset;
          if (edata.Nodes != null) {
            for (var i = 0; i < edata.Nodes.Length; i++) {
              edata.Nodes[i] += offset;
            }
          }
        }

        foreach (var edata in room.Triggers) {
          edata.Position += offset;
          if (edata.Nodes != null) {
            for (var i = 0; i < edata.Nodes.Length; i++) {
              edata.Nodes[i] += offset;
            }
          }
        }

        Logger.Log(LogLevel.Debug, "bitsbolts.Hooks.GlobalRoomLoader", $"Offsetting room {room.Name} entities by {offset}");

        dd.Set("_bitsbolts_globalroomoffset", (bool?) true);
      }
    }


    // fixme: we need to fix the entityid of the globally loaded entities so that
    // they match up with the global room their are from (e.g. so it is
    // consistent with EntityData.Level), even though they being loaded in a
    // completely seperate level.
    // fixme: similarly, we need to set the correct levelOffset for global
    // entities loaded this way
    // fixme: and we need fix both of these in LoadCustomEntity too...
    // ...which means an everest hook :(
    //
    // td;lr: https://github.com/EverestAPI/Everest/pull/689 fixes all these
    // problems with LoadVanillaEntity and adds optional params to
    // LoadCustomEntity.
    //
    // At the current moment, this loader "works" well enough to be usable for
    // positionally-independent entities (e.g. controllers, global triggers,
    // triggers that are resized to screen size), but breaks entities loaded via
    // this class that have a position in the world or those that use DNL flags.
    // At some point in the future when the above PR is stable, we'll refactor
    // all of this garbage IL code into a simple On hook that iteratively calls
    // LevelExt.LoadEntity for every entity in CachedRooms.
    // Right now, there's no effort made to fix levelOffset or EntityData.Level
    // because it'll result in inconsistent behavior: it'll only fix some of the
    // vanilla entities that are not already being patched by everest in
    // LoadCustomEntity, and it won't fix any custom entities.


    // private delegate String ModEntityIDLevelOffset(String levelName, EntityData entity, ref Vector2 levelOffset);
    private static void ILLevelLoadLevel(ILContext il) {
      var cursor = new ILCursor(il);

      // // easier to bind to this local at it's first use instead at definition
      // int localLevelOffset = 0;
      // if (!cursor.TryGotoNext(
      //   MoveType.After,
      //   instr => instr.MatchLdloc(out localLevelOffset),
      //   instr => instr.MatchNewobj<Checkpoint>()
      // )) {
      //   Logger.Log(LogLevel.Error, "bitsbolts.Hooks.GlobalRoomLoader", $"IL@{cursor.Index}: Hook failed to bind to local `vector` (i.e. levelOffset) on Checkpoint::ctor in Level.LoadLevel");
      //   return;
      // }

      // // reset our IL cursor to start from the top again
      // cursor = new ILCursor(il);


      // Hook the entity loop:
      // foreach (EntityData entityData in levelData.Entities) {
      if (!cursor.TryGotoNext(
        MoveType.After,
        instr => instr.MatchLdfld<LevelData>("Entities")  // type: List<EntityData>
      )) {
        Logger.Log(LogLevel.Error, "bitsbolts.Hooks.GlobalRoomLoader", $"IL@{cursor.Index}: Hook failed to find foreach (EntityData entityData in levelData.Entities) in Level.LoadLevel");
        return;
      }

      // The hook to hotswap levelData.Entities, and also do some level
      // processing and caching for later use.
      cursor.Emit(OpCodes.Ldarg_0);
      cursor.Emit(OpCodes.Ldarg_1);
      cursor.EmitDelegate<Func<List<EntityData>, Level, Player.IntroTypes, List<EntityData>>>(
        (List<EntityData> entities, Level level, Player.IntroTypes intro) => {
          GlobalRoomLoader.RoomIter = null;

          var levelControllers = DynamicData.For(level.Session.MapData)
            .Get<Dictionary<LevelData, GlobalRoomLoaderController>>("_bitsbolts_globalroom_levelcontrollers");

          if (levelControllers.Count == 0) { return entities; }

          Logger.Log(LogLevel.Debug, "bitsbolts.Hooks.GlobalRoomLoader", $"Handling {levelControllers.Count} global controllers");

          GlobalRoomLoader.RoomIter = levelControllers.Where(pair => {
            var controller = pair.Value;
            return level.Session.GetFlagstrs(controller.Flags).All(f => f.Value)
              && (controller.Rooms == null || controller.Rooms.ContainsKey(level.Session.Level))
              && ((intro != Player.IntroTypes.Transition) || controller.DuringTransition);
          }).Select(pair => pair.Key);  // type: LevelData


          if (!GlobalRoomLoader.RoomIter.Any()) {
            return entities;
          }


          Logger.Log(LogLevel.Debug, "bitsbolts.Hooks.GlobalRoomLoader", $"Live concating entities!");
          EntityListBuffer.Clear();
          EntityListBuffer.AddRange(entities);
          foreach (LevelData room in GlobalRoomLoader.RoomIter) {
            EntityListBuffer.AddRange(room.Entities);
          }
          return EntityListBuffer;
        }
      );


      // Don't bother doing this...

      // // Literally just fetch the localEntityData local in the inner loop.
      // int localEntityData = 0;
      // if (!cursor.TryGotoNext(
      //   MoveType.After,
      //   // instr => instr.MatchCall(new List<Celeste.EntityData>().GetEnumerator().GetType(), "get_Current"),
      //   instr => instr.OpCode == OpCodes.Call && instr.Operand is MethodReference method
      //     && method.Name == "get_Current" && method.DeclaringType.FullName == "System.Collections.Generic.List`1/Enumerator<Celeste.EntityData>",
      //   instr => instr.MatchStloc(out localEntityData)
      // )) {
      //   Logger.Log(LogLevel.Error, "bitsbolts.Hooks.GlobalRoomLoader", $"IL@{cursor.Index}: Hook failed to bind to entityData = List<EntityData>.Enumerator.get_Current() after foreach (EntityData entityData in levelData.Entities) in Level.LoadLevel");
      //   return;
      // }


      // // Hook the inner
      // if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchLdfld<Celeste.LevelData>("Name"))) {
      //   Logger.Log(LogLevel.Error, "bitsbolts.Hooks.GlobalRoomLoader", $"IL@{cursor.Index}: Hook failed to find ldfld Celeste.LevelData::Name to mod EntityID.ctor after foreach (EntityData entityData in levelData.Entities) in Level.LoadLevel");
      //   return;
      // }

      // cursor.Emit(OpCodes.Ldloc_S, (byte) localEntityData);
      // cursor.Emit(OpCodes.Ldloca_S, (byte) localLevelOffset);
      // cursor.EmitDelegate<ModEntityIDLevelOffset>(
      //   (String levelName, EntityData entity, ref Vector2 levelOffset) => {
      //     if (levelName != entity.Level.Name) {
      //       levelName = entity.Level.Name;
      //       levelOffset = new Vector2(entity.Level.Bounds.Left, entity.Level.Bounds.Top);
      //     }
      //     return levelName;
      //   }
      // );


      if (!cursor.TryGotoNext(
        MoveType.After,
        instr => instr.MatchLdfld<LevelData>("Triggers")  // type: List<EntityData>
      )) {
        Logger.Log(LogLevel.Error, "bitsbolts.Hooks.GlobalRoomLoader", $"IL@{cursor.Index}: Hook failed to find foreach (EntityData entityData in levelData.Triggers) in Level.LoadLevel");
        return;
      }

      cursor.EmitDelegate<Func<List<EntityData>, List<EntityData>>>(
        (List<EntityData> triggers) => {
          if (GlobalRoomLoader.RoomIter == null) { return triggers; }

          Logger.Log(LogLevel.Debug, "bitsbolts.Hooks.GlobalRoomLoader", $"Live concating triggers!");
          EntityListBuffer.Clear();
          EntityListBuffer.AddRange(triggers);
          foreach (LevelData room in GlobalRoomLoader.RoomIter) {
            EntityListBuffer.AddRange(room.Triggers);
          }

          GlobalRoomLoader.RoomIter = null;
          return EntityListBuffer;
        }
      );
    }
  }
}
