using System;
using System.Linq;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts {
  internal static class LevelExt {
    [OnLoad]
    public static void Load() {
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;
      On.Monocle.Tracker.Initialize += TrackerInitialize;
    }

    [OnUnload]
    public static void Unload() {
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
      On.Monocle.Tracker.Initialize -= TrackerInitialize;
    }

    private static void TrackerInitialize(On.Monocle.Tracker.orig_Initialize orig) {
      orig();
      MonocleExt.TrackComponent<Leader>(inherited: true);
    }


    public static void TeleportToExact(this Level level, Player player, string nextLevel, Player.IntroTypes introType, Vector2? target = null) {
      Leader.StoreStrawberries(player.Leader);
      Vector2 position = player.Position;
      level.Remove(player);
      level.UnloadLevel();
      level.Session.Level = nextLevel;
      level.Session.RespawnPoint = target != null ? target : level.GetSpawnPoint(Vector2.Zero);
      if (introType == Player.IntroTypes.Transition) {
        player.Position = level.Session.RespawnPoint.Value;
        player.Hair.MoveHairBy(player.Position - position);
        player.MuffleLanding = true;
        level.Add(player);
        level.LoadLevel(Player.IntroTypes.Transition, false);
        level.Entities.UpdateLists();
      } else {
        level.LoadLevel(introType, false);
        level.Entities.UpdateLists();
        player = level.Tracker.GetEntity<Player>();
      }
      level.Camera.Position = player.CameraTarget;
      level.Update();
      Leader.RestoreStrawberries(player.Leader);
    }


    // todo: this really shouldn't be static... but then again, backdrops aren't
    // shared between Levels so does it really matter?
    private static Dictionary<Backdrop, Vector2> BackdropOffset = new Dictionary<Backdrop, Vector2>();

    internal static void TeleportSeamlessly(this Level level, Player player, Vector2 delta) {
      level.Camera.Position += delta;

      // teleport the player...
      player.Position += delta;
      player.Hair.MoveHairBy(delta);

      // fixme: fuck this only needs to be applied on all the teleported actors!!!
      // ...and all leaders (including the one on Player too)
      foreach (Leader leader in level.Tracker.GetComponents<Leader>()) {
        for (var i = 0; i < leader.PastPoints.Count; i++) {
          leader.PastPoints[i] += delta;
        }

        foreach (var follower in leader.Followers) {
          if (follower != null) {
            follower.Entity.Position += delta;
          }
        }
      }

      // Add a temporary offset to looping backdrops that is reset when a
      // non-transition level load occurs
      foreach (Backdrop backdrop in level.Background.Backdrops.Concat(level.Foreground.Backdrops)) {
        var offset = Vector2.Zero;

        if (backdrop.LoopX) {
          backdrop.Position.X += (offset.X = delta.X * backdrop.Scroll.X);
        }

        if (backdrop.LoopY) {
          backdrop.Position.Y += (offset.Y = delta.Y * backdrop.Scroll.Y);
        }

        if (offset != Vector2.Zero) {
          BackdropOffset[backdrop] = offset;
        }
      }
    }

    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level, Player.IntroTypes intro, bool isFromLoader
    ) {
      // If the backdrop is still in the level on next load, then we can take
      // time to undo our offset changes (since backdrop is only reinitialized
      // by LevelLoader, so our changes are otherwise permanent)
      if (intro != Player.IntroTypes.Transition) {
        foreach (var kvp in BackdropOffset) {
          Backdrop backdrop = kvp.Key;
          Vector2 offset = kvp.Value;
          backdrop.Position -= offset;
        }
        BackdropOffset.Clear();
      }
      orig(level, intro, isFromLoader);
    }
  }
}
