using System;
using System.Linq;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;


namespace Bitsbolts {
  internal class MapDataTracker {
    [OnLoad]
    // Avoid confusion with the instance method Load
    private static void ModuleLoad() {
      // Manual hook in Bitsbolts.Events:
      // Bitsbolts.Events.OnMapDataLoad += OnMapDataLoad;

      MapDataTracker.Initialize();
    }


    internal static HashSet<String> TrackedEntityDataNames = new HashSet<String>();

    internal Dictionary<String, List<EntityData>> EntityDatas = new Dictionary<String, List<EntityData>>();

    internal MapDataTracker() {}

    internal static void Initialize() {
      foreach (Type type in typeof(MapDataTracker).Assembly.GetTypesSafe()) {
        foreach (MapDataTracked attribute in type.GetCustomAttributes(typeof(MapDataTracked), false)) {
          MapDataTracker.TrackEntityName(attribute.Name);
        }
      }
    }

    internal void Load(MapData mapData) {
      foreach (var dataName in TrackedEntityDataNames) {
        if (!this.EntityDatas.ContainsKey(dataName)) {
          this.EntityDatas[dataName] = new List<EntityData>();
        } else {
          this.EntityDatas[dataName].Clear();
        }
      }

      foreach (var data in mapData.Levels.SelectMany(l => l.Entities)) {
        if (TrackedEntityDataNames.Contains(data.Name)) {
          this.EntityDatas[data.Name].Add(data);
        }
      }
    }

    internal EntityData GetEntityData(String entityName) {
      var list = this.EntityDatas[entityName];
      return list.Count == 0 ? null : list[0];
    }

    internal List<EntityData> GetEntityDatas(String entityName) {
      return this.EntityDatas[entityName];
    }

    internal static void TrackEntityName(String name) {
      TrackedEntityDataNames.Add(name);
    }


    // Needs to be internal so this is accessible from Bitsbolts.Events.
    internal static void OnMapDataLoad(MapData mapData) {
      mapData.GetTracker().Load(mapData);
    }
  }

  internal static class MapDataTrackerExt {
    public static MapDataTracker GetTracker(this MapData mapData) {
      var dd = DynamicData.For(mapData);
      if (dd.TryGet("Tracker", out Object tracker) && tracker is MapDataTracker tracker2) {
        return tracker2;
      } else {
        var tracker3 = new MapDataTracker();
        dd.Set("Tracker", tracker3);
        return tracker3;
      }
    }

    public static bool HasTracker(this MapData mapData) {
      var dd = DynamicData.For(mapData);
      return dd.TryGet("Tracker", out Object tracker) && tracker is MapDataTracker;
    }
  }


  [AttributeUsage(AttributeTargets.Class, AllowMultiple = false, Inherited = false)]
  internal class MapDataTracked : Attribute {
    public String Name;
    internal MapDataTracked(string name) { this.Name = name; }
  }
}
