using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod.Entities;

using Flagstr = System.Tuple<string, bool>;
using Flagstrs = System.Collections.Generic.Dictionary<string, bool>;


namespace Bitsbolts {
  internal static class Parsing {
    internal class ParseException : Exception {
      internal ParseException() {}
      internal ParseException(string message) : base(message) {}
      internal ParseException(string message, Exception inner) : base(message, inner) {}
    }


    internal static EntityID ParseEID(this string input) {
      try {
        var eid = new EntityID("", 0);
        eid.Key = input;
        return eid;
      } catch (IndexOutOfRangeException ex) {
        throw new ParseException($"Parsing EntityID encountered invalid input: {input}", ex).Carded();
      }
    }

    internal static List<EntityID> ParseEIDs(this string input) {
      return input
        .Split(new char[] { ',' }, System.StringSplitOptions.RemoveEmptyEntries)
        .Select(Parsing.ParseEID)
        .ToList();
    }


    internal static (string Flag, bool Enable) ParseFlagstr(
      this string input, bool allowInversions = true
    ) {
      if (input.Length >= 1 && input[0] == '!') {
        if (!allowInversions) {
          throw new ParseException($"Flag {input} cannot be inverted in this context!").Carded();
        }
        return (input.Substring(1), false);
      }
      return (input, true);
    }

    internal static Dictionary<string, bool> ParseFlagstrs(
      this string input, bool allowInversions = true
    ) {
      return input
        .Split(new char[] { ',' }, System.StringSplitOptions.RemoveEmptyEntries)
        .Select(str => Parsing.ParseFlagstr(str, allowInversions))
        .ToDictionary(flag => flag.Flag, flag => flag.Enable);
    }

    internal static Flagstrs InvertFlagstrs(this Flagstrs flags) {
      var output = new Flagstrs(flags);
      foreach (var flag in flags) {
        output[flag.Key] = !flag.Value;
      }
      return output;
    }

    internal static void SetFlagstrs(this Session session, Dictionary<string, bool> flags) {
      foreach (var flag in flags) {
        session.SetFlag(flag.Key, flag.Value);
      }
    }

    internal static Dictionary<string, bool> GetFlagstrs(this Session session, Dictionary<string, bool> flags) {
      var output = new Flagstrs(flags);
      foreach (var flag in flags) {
        output[flag.Key] = session.GetFlag(flag.Key) == flag.Value;
      }
      return output;
    }


    internal static string ParseGlobs(this string input) {
      return input
        .Split(new char[] { ',' }, System.StringSplitOptions.RemoveEmptyEntries)
        .Globs();
    }

    internal static (string PositiveRegex, string NegativeRegex) ParseGlobsWithNegatives(this string input) {
      var globs = new List<string>();
      var invertedGlobs = new List<string>();

      foreach (
        var glob in input.Split(new char[] { ',' }, System.StringSplitOptions.RemoveEmptyEntries)
      ) {
        if (glob.Length >= 1 && glob[0] == '!') {
          invertedGlobs.Add(glob.Substring(1));
        } else {
          globs.Add(glob);
        }
      }

      return (Parsing.Globs(globs), Parsing.Globs(invertedGlobs));
    }

    internal static string Glob(this string pattern) {
      return "^" + Regex.Escape(pattern).Replace("\\*", ".*") + "$";
    }

    internal static string Globs(this IEnumerator<string> patterns) {
      return Parsing.Globs(patterns.Enumerable());
    }

    internal static string Globs(this IEnumerable<string> patterns) {
      string regex = "^(";
      foreach (var pattern in patterns) {
        regex += (regex.Length == 2 ? "" : ")|(")
          + Regex.Escape(pattern).Replace("\\*", ".*");
      }
      return regex + ")$";
    }


    internal static Dictionary<string, LevelData> ParseLevelsList(
      this string input, MapData mapData, bool allowInversions = true
    ) {
      var matched = new Dictionary<string, LevelData>();
      var (regex, negativeRegex) = input.ParseGlobsWithNegatives();
      foreach (var levelData in mapData.Levels) {
        if (
          Regex.IsMatch(levelData.Name, regex)
          && !Regex.IsMatch(levelData.Name, negativeRegex)
        ) {
          matched[levelData.Name] = levelData;
        }
      }

      return matched;
    }
  }
}
