using System;

using Microsoft.Xna.Framework;
using Monocle;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using MonoMod.RuntimeDetour;
using MonoMod.Utils;
using Celeste;
using Celeste.Mod;
using Celeste.Mod.Entities;

using Bitsbolts.Components;


namespace Bitsbolts.Triggers {
  [CustomEntity("Bitsbolts/ChangeRespawnCrossRoom")]
  public class ChangeRespawnCrossRoom : ChangeRespawnTrigger {
    [OnLoad]
    public static void Load() {
      IL.Celeste.ChangeRespawnTrigger.OnEnter += ILChangeRespawnTriggerOnEnter;
      IL.Celeste.PlayerDeadBody.End += ILPlayerDeadBodyEnd;
      On.Celeste.Level.LoadLevel += OnLevelLoadLevel;
      On.Celeste.LevelExit.Begin += OnLevelExitBegin;
    }

    [OnUnload]
    public static void Unload() {
      IL.Celeste.ChangeRespawnTrigger.OnEnter -= ILChangeRespawnTriggerOnEnter;
      IL.Celeste.PlayerDeadBody.End -= ILPlayerDeadBodyEnd;
      On.Celeste.Level.LoadLevel -= OnLevelLoadLevel;
      On.Celeste.LevelExit.Begin -= OnLevelExitBegin;
    }


    public readonly bool Exact = false;
    public ChangeRespawnCrossRoom(EntityData data, Vector2 offset) : base(data, offset) {
      this.Exact = data.Bool("exact", false);

      var flags = data.Attr("flags", "").ParseFlagstrs();
      if (flags.Count > 0) { base.Add(new FlagstrTrigger(flags)); }
      base.Add(new TransitionTrigger(before: false));
    }

    public override void Added(Scene scene) {
      // Skip ChangeRespawnTrigger.Added in the base.Added chain
      // We can't have it changing our target for us... that's our job
      var EntityAdded = (Action<Scene>) Activator.CreateInstance(
        typeof(Action<Scene>),
        this,
        typeof(Entity).GetMethod("Added").MethodHandle.GetFunctionPointer()
      );
      EntityAdded(scene);

      var levelData = (scene as Level).Session.MapData.GetAt(this.Target);
      if (levelData == null) {
        throw new Exception("ChangeRespawnCrossRoom respawn target is out of bounds of all rooms!");
      }

      if (!this.Exact) {
        base.Target = levelData.Spawns.ClosestTo(this.Target);
      }
    }

    static private void ILChangeRespawnTriggerOnEnter(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.After, instr => instr.MatchStfld<Session>("RespawnPoint"))) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.ChangeRespawnCrossRoom", $"IL@{cursor.Index}: Hook failed to find stfld session.RespawnPoint = ... in ChangeRespawnTrigger.OnEnter");
        return;
      }

      Logger.Log("Bitsbolts.Entities.ChangeRespawnCrossRoom", $"IL@{cursor.Index}: Adding delegate in ChangeRespawnTrigger.OnEnter to mod after session.RespawnPoint = ...");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Action<ChangeRespawnTrigger>>(
        (trigger) => {
          if (trigger is ChangeRespawnCrossRoom trigger2) {
            var levelData = (trigger2.Scene as Level).Session.MapData.GetAt(trigger2.Target);
            MainModule.Session.RespawnPointLevel = levelData.Name;
          } else {
            // if the respawn point is set by means other than this trigger, we
            // need to unset the respawn level
            MainModule.Session.RespawnPointLevel = null;
          }
        }
      );
    }

    static private void ILPlayerDeadBodyEnd(ILContext il) {
      var cursor = new ILCursor(il);

      if (!cursor.TryGotoNext(MoveType.Before, instr => instr.MatchStfld<PlayerDeadBody>("DeathAction"))) {
        Logger.Log(LogLevel.Error, "Bitsbolts.Entities.ChangeRespawnCrossRoom", $"IL@{cursor.Index}: Hook failed to find stfld this.DeathAction = ... in PlayerDeadBody.End");
        return;
      }

      Logger.Log("Bitsbolts.Entities.ChangeRespawnCrossRoom", $"IL@{cursor.Index}: Adding delegate in PlayerDeadBody.End to mod this.DeathAction = ...");

      cursor.Emit(OpCodes.Ldarg_0);
      cursor.EmitDelegate<Func<Action, PlayerDeadBody, Action>>(
        (orig, body) => {
          var level = body.Scene as Level;
          if (level.Session.RespawnPoint == null || MainModule.Session.RespawnPointLevel == null) {
            MainModule.Session.RespawnPointLevel = null;
            return orig;
          }

          return () => {
            var newLevel = MainModule.Session.RespawnPointLevel;
            MainModule.Session.RespawnPointLevel = null;
            level.TeleportToExact(
              player: DynamicData.For(body).Get<Player>("player"),
              nextLevel: newLevel,
              introType: Player.IntroTypes.Respawn,
              target: level.Session.RespawnPoint
            );
          };
        }
      );
    }

    private static void OnLevelLoadLevel(
      On.Celeste.Level.orig_LoadLevel orig, Level level,
      Player.IntroTypes playerIntro, bool isFromLoader
    ) {
      if (MainModule.Session != null && MainModule.Session.RespawnPointLevel != null) {
        var newLevel = MainModule.Session.RespawnPointLevel;
        MainModule.Session.RespawnPointLevel = null;
        if (
          (level.Session.RespawnPoint != null)
          && (playerIntro != Player.IntroTypes.Transition)
          // && (playerIntro != Player.IntroTypes.Respawn || isFromLoader)
          && (level.Session.Level != newLevel)
        ) {
          level.Session.Level = newLevel;
        }
      }

      orig(level, playerIntro, isFromLoader);
    }

    private static void OnLevelExitBegin(
      On.Celeste.LevelExit.orig_Begin orig, LevelExit exit
    ) {
      var dd = DynamicData.For(exit);
      switch (dd.Get<LevelExit.Mode>("mode")) {
        case LevelExit.Mode.GiveUp:
        case LevelExit.Mode.Restart:
        case LevelExit.Mode.GoldenBerryRestart:
          MainModule.Session.RespawnPointLevel = null;
          break;
      }

      orig(exit);
    }
  }
}
